import glob
import re
import sys

# Parameter File Setup
######################
# required variables for this stage
# it is an error if these aren't defined.
req_params_file_vars = [ "optimization_iterations", "number_of_particles" ]

# this is the list of variables that might be defined that we should write 
# to the params file if they are defined AND the conditional expression 
# evaluates to True.
# the format of this data structure is
# [ [ parameter, conditional_expression ], ... ]
opt_params_file_vars =  [ ["starting_regularization", "True"],
			  ["ending_regularization", "True"],
			  ["checkpointing_interval", "do_checkpointing"],
			  ["procrustes_interval", "do_procrustes"],
			  ["procrustes_scaling", "do_procrustes"],
			  ["optimization_iterations_completed", "True"]
]

# Commandline Setup
######################
# commandline options
# the format of this data structure is
# [ [command, [ enabling_expression ], [ dependencies ] ]
# So the command will be written to the commandline IF the 
# enabling expression is True AND the dependencies are defined.
# If the dependencies are not defined, we may print a warning message
# and not put the param to the commandline or alternately throw an error.
commandline_params = [ ]

global command_name
if sys.platform == "win32":
	command_name = "ShapeWorksRun.exe"
else:
	command_name = "ShapeWorksRun"

Optimize_Output_Prefix = "Optimize"

def copy_stage_inputs( output_path ) :
	# Technically, the inputs to this stage are 
	# the image outputs from the Filter stage
	copy_outputs_from_stage( "Filter", output_path )
	if ( not( reuse_previous_results ) ) :
		copy_outputs_from_stage( "Initialize", output_path )
	else :
		copy_outputs_from_stage( "Optimize", output_path )

def build_stage_outputs_list( ) :

	global shapeworks_max_iter_completed
	shapeworks_max_iter_completed = 0
	global shapeworks_done
	shapeworks_done = False
	# gather the outputs
	# this uses the global project_path and working_dir_path
	# ... should probably be redesigned :(
	
	# first look to see if the "done" file exists.
	done_file_path = os.path.join( working_dir_path, "done" )
	
	is_checkpoint = True
	
	if ( os.path.exists( done_file_path ) ) :
		is_checkpoint = False
		# the outputs will have the Optimize prefix and will
		# be located in the root of the working dir
		outputs = glob.glob( os.path.join( working_dir_path, 
			Optimize_Output_Prefix + '*' ) )
			
		shapeworks_done = True
		shapeworks_max_iter_completed = optimization_iterations
		
	else :
		is_checkpoint = True
		# when a checkpoint is written out, a directory called
		# iter[n] is written out, where n is the total number
		# of iterations completed.
		checkpoint_dirs = glob.glob( os.path.join( working_dir_path,
			'iter*' ) )
		
		# find the maximum iteration completed.
		shapeworks_max_iter_completed = 0
		for checkpoint_dirname in checkpoint_dirs :
			
			iter_match = re.search( "iter(\d*)", checkpoint_dirname )
			if ( iter_match != None ) :
				iter_num = int( iter_match.groups()[0] )
				if ( iter_num > shapeworks_max_iter_completed ) :
					shapeworks_max_iter_completed = iter_num
		
		# Now grab all the files matching that iteration number	
		outputs = glob.glob( os.path.join( working_dir_path, 
			'iter' + str( shapeworks_max_iter_completed ) + '/*' ) )
			
	# as a final step, remove the path in front of the filename
	# to make it consistent with the other stages.
	outputs_tmp = []
	for path in outputs :
		outputs_tmp.append( os.path.basename( path ) )
		
	outputs = outputs_tmp
	return outputs

def build_output_path( ) :
	if ( shapeworks_done ) :
		return working_dir_path
	else :
		return os.path.join( working_dir_path, 'iter'+str(shapeworks_max_iter_completed) )

def build_params_list( ) :

	global inputs
	global point_files
	global transform_file
	global output_transform_file
	global output_points_prefix
	global keep_checkpoints

	req_params_file_vars.append( "inputs" )
	inputs = import_outputs_list( "Filter" )

	req_params_file_vars.append( "point_files" )
	if ( not( reuse_previous_results ) ) :
		outputs = import_outputs_list( "Initialize" )
	else :
		outputs = import_outputs_list( "Optimize" )

	lpts_regex = re.compile(".*\.lpts$");
	# the outputs will contain all the files output by the previous stage
	# however, we just want the .lpts files.
	point_files = list( filter( lpts_regex.search, outputs ) )

	req_params_file_vars.append( "output_points_prefix" )
	output_points_prefix = "Optimize"	
	
	trans_regex = re.compile( ".*_transform.*" )

	try:
		transform_file = list( filter( trans_regex.search, outputs ) )[0]		
	except:
		transform_file = ""

	if ( transform_file != "" ) :
		req_params_file_vars.append( "transform_file" )
	
	req_params_file_vars.append( "output_transform_file" )
	output_transform_file = Optimize_Output_Prefix + "_transform"
	
	req_params_file_vars.append( "keep_checkpoints" )
	keep_checkpoints = True
	
def update_state_stage_specific( file ) :

	if ( shapeworks_done ) :
		file.write( "is_running = False\n" )
		file.write( "percent_complete = 100.0\n" )
	else :
		file.write("is_running = True\n")
		
#		print("%f %f" % (shapeworks_max_iter_completed,optimization_iterations))
		
		percent_complete = 100.0*float(shapeworks_max_iter_completed) / float(optimization_iterations)
		
		file.write( "percent_complete = %f\n" % percent_complete )

	file.write("optimization_iterations_completed = %s\n" % shapeworks_max_iter_completed )

def cleanup( ) :

	# if Optimize is done, we don't need to do any cleanup, because
	# the directory is about to be deleted.
	
	if ( not( shapeworks_done ) ) :
	# find all the files whose iteration numbers are less than the one 
	# we have been using thus far
		# when a checkpoint is written out, a directory called
		# iter[n] is written out, where n is the total number
		# of iterations completed.
		
		iter_dirs = glob.glob( os.path.join( working_dir_path,
			'iter*' ) )
				
		# remove the iter files from iterations less than the max completed.
		for iter_dir in iter_dirs :
		
			iter_match = re.search( "iter(\d*)", iter_dir )
			if ( iter_match != None ) :
				iter_num = int( iter_match.groups()[0] )
				
				if ( iter_num <= shapeworks_max_iter_completed ) :
					# remove the iteration file
					shutil.rmtree( iter_dir )
