#!/usr/bin/env python
#  
#  For more information, please see: http://software.sci.utah.edu
#  
#  The MIT License
#  
#  Copyright (c) 2011 Scientific Computing and Imaging Institute,
#  University of Utah.
#  
#  
#  Permission is hereby granted, free of charge, to any person obtaining a
#  copy of this software and associated documentation files (the "Software"),
#  to deal in the Software without restriction, including without limitation
#  the rights to use, copy, modify, merge, publish, distribute, sublicense,
#  and/or sell copies of the Software, and to permit persons to whom the
#  Software is furnished to do so, subject to the following conditions:
#  
#  The above copyright notice and this permission notice shall be included
#  in all copies or substantial portions of the Software.
#  
#  THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
#  OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
#  FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
#  THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
#  LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
#  FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
#  DEALINGS IN THE SOFTWARE.
#  

from sys import argv
from sys import exit
import os
import shutil
import sys
import traceback

# Copy named files to dest
def copy_files( srcs, dest ) :
	
	for file_name in srcs :
		try:
			shutil.copy( file_name, dest )
		except:
			print( 'Could not copy %s to %s' % ( filename, dest ) )
			exit( 1 )

# write out the state.txt file to the target directory.
def update_state( state_fname ) :
	
	state_file = open( state_fname, 'w' )
	
	# write empty outputs
	state_file.write( "outputs = []\n" )
	
	# set percent complete to 0
	state_file.write( "percent_complete = 0.0\n" )

	# set is_running to true
	state_file.write( "is_running = True\n" )
	
	state_file.close()


# grab the outputs from the state of stage "stage_id"
def import_outputs_list( stage_id ) :

	# special case the first stage -- the inputs to the second
	# stage will be the inputs specified in the params.txt
	if ( stage_id == "Files" ) :
		return inputs

	inputs_stage_dir = os.path.join( project_path, stage_id )
	state_fname = os.path.join( inputs_stage_dir, "state.txt" )
	
	if ( not( os.path.exists( state_fname ) ) ) :
		print( 'No such file %s.' % state_fname )
		exit( 1 )
	
	try:
		state_file = open( state_fname, 'r' ).read()
	except:
		print( 'Failed to open %s\n' % state_fname )
		exit( 1 )
	
	try:
		exec( state_file, globals() )
	except:
		print( 'Failed to exec %s in prepare.py' % state_fname )
		traceback.print_exc()
		exit( 1 )
	
	# we expect that outputs was defined in the state.txt file
	# it should list all the output files from the previous stage.
	try:
		outputs
	except NameError:
		print( 'outputs not defined in %s\n' % state_fname )
		exit( 1 )
	
	return outputs


def build_stage_outputs_from_inputs( stage_id, inputs ) :
	outputs = []
	for fname in inputs :
		(base, suffix) = os.path.splitext( fname )
		outputs.append(base+stage_id+suffix)
		
	return outputs


# copy the outputs from the specified stage into the
# output directory.
def copy_outputs_from_stage( from_stageid, output_dir ) :

	inputs_stage_dir = os.path.join( project_path, from_stageid )

	outputs = import_outputs_list( from_stageid )
	
	inputs_full = []

	for file_name in outputs :
		if ( os.path.isabs( file_name ) ) :
			full_file_name = file_name 
		else :
			full_file_name = os.path.join( inputs_stage_dir, file_name )
		if ( not( os.path.isfile( full_file_name ) ) ) :
			print( 'File %s does not exist!' % full_file_name )
			exit( 1 )	
		inputs_full.append( full_file_name )

	copy_files( inputs_full, output_dir )

# write a list in the params file format
def write_list_param( file, param_name, param_val ) :
	file.write( "(%s\n" % param_name )
	
	for item in param_val :
		if ( isinstance( item,str ) ) :
			file.write( '"%s"\n' % item )
		else :
			file.write( '%s\n' % item )
	
	file.write( ")\n" )

# write out a scalar value in the params file format
def write_scalar_param( file, param_name, param_val ) :

	scalar_val = param_val
	
	if ( param_val is True ) :
		scalar_val = 1
	if ( param_val is False ) :
		scalar_val = 0
	if ( isinstance( scalar_val, str ) ) :
		file.write( '(%s "%s")\n' % ( param_name, scalar_val ) )
	else :
		file.write( '(%s %s)\n' % ( param_name, scalar_val ) )

# write a params file parameter to a file
def write_param( file, param_name ) :

	try:
		param_val = eval( param_name )
		if ( type( param_val ) is list ) :
			write_list_param( file, param_name, param_val )
		else :
			write_scalar_param( file, param_name, param_val )
	except NameError:
		# no problem here, we just return.
		return

# write out the shapeworks inputs as a .params file
def write_shapeworks_params( fname ) :
	# Check for the existence of the necessary variables.  If they are
	# not there, throw an exception :)

	try:
		params_f = open( fname, 'w' )
	except:
		print( 'Unable to open output %s' % fname )
		exit( 1 )


	# first handle the required parameters for this stage
	for var_name in req_params_file_vars :
		try:
			eval( var_name )
		except NameError:
			print( 'Required variable %s is not defined\n' % var_name )
			exit( 1 )
		
	for var_name in req_params_file_vars :
		write_param( params_f, var_name )
	
	# now handle the optional parameters......
	for elem in opt_params_file_vars :
		(var_name, cond_exp) = elem
	
		try: 	
			if ( eval( cond_exp ) ) :
				write_param( params_f, var_name )
		except NameError:
			continue
			
	params_f.close()


# write a list in the XML file format
def write_list_param_xml( file, param_name, param_val ) :
	file.write( "<%s>\n" % param_name )
	
	for item in param_val :
		file.write( '%s\n' % item )
	
	file.write( "</%s>\n" % param_name )

# write out a scalar value in the XML file format
def write_scalar_param_xml( file, param_name, param_val ) :

	scalar_val = param_val
	
	if ( param_val is True ) :
		scalar_val = 1
	if ( param_val is False ) :
		scalar_val = 0
	file.write( '<%s> %s </%s>\n' % ( param_name, scalar_val, param_name ) )

# write a XML file parameter to a file
def write_param_xml( file, param_name ) :

	try:
		param_val = eval( param_name )
		if ( type( param_val ) is list ) :
			write_list_param_xml( file, param_name, param_val )
		else :
			write_scalar_param_xml( file, param_name, param_val )
	except NameError:
		# no problem here, we just return.
		return


# write out the shapeworks inputs as an XML file
def write_shapeworks_xml( fname ) :
	# Check for the existence of the necessary variables.  If they are
	# not there, throw an exception :)

	try:
		xml_f = open( fname, 'w' )
	except:
		print( 'Unable to open output %s' % fname )
		exit( 1 )

	xml_f.write( '<?xml version="1.0" ?>\n' )

	# first handle the required parameters for this stage
	for var_name in req_params_file_vars :
		try:
			eval( var_name )
		except NameError:
			print( 'Required variable %s is not defined\n' % var_name )
			exit( 1 )
		
	for var_name in req_params_file_vars :
		write_param_xml( xml_f, var_name )
	
	# now handle the optional parameters......
	for elem in opt_params_file_vars :
		(var_name, cond_exp) = elem
	
		try: 	
			if ( eval( cond_exp ) ) :
				write_param_xml( xml_f, var_name )
		except NameError:
			continue
			
	xml_f.close()

if __name__ == "__main__" :

	if ( len( argv ) > 1 and argv[1] == '-x' ) :
		use_xml = True
		argv.pop(1)
	else :
		use_xml = False

	if len( argv ) != 4 :
			print( 'Usage: %s [-x] project_path working_dir_path stage_id.' % argv[0] )
			exit( 1 )
	if ( not( os.path.isdir( argv[1] ) ) ) :
		print( 'Project path is not a directory: %s.' % argv[1] )
		exit( 1 )
	if ( not( os.path.isdir( argv[2] ) ) ) :
		print( 'Output path is not a directory: %s.' % argv[2] )
		exit( 1 )

	script_name = argv[0]
	scripts_path = os.path.dirname( script_name ) 

	# check for the existence of a parameter file in the input directory.
	project_path = argv[1]
	
	parameter_fname = os.path.join(project_path,"params.txt")
	if ( not( os.path.exists( parameter_fname ) ) ) :
		print( 'No such file %s.' % parameter_fname )
		exit( 1 )

	working_dir_path = argv[2]
	current_stage_id = argv[3]

	# copy the parameter file to the output directory
	copy_files( [ parameter_fname ], working_dir_path )

	# source the parameter.txt file to read in the variables
	# note: this is a security issue if the origin of this file is not 
	# trusted.
	try:
		parameter_file = open( parameter_fname ).read()
	except:
		print( 'Could not open %s for reading.' % parameter_fname )  
		exit( 1 )

	# Run the params.txt file so that its variables get read in.
	# If this fails, the following line will return an exception.
	# We choose not to handle it, as the commandline error is more
	# informative this way.
	exec( parameter_file )
	
	current_stage_index = 0
	
	valid_stage = False
		
	# find the index of the current stage.
	try:
		for stage_id in stage_ids :
			if ( stage_id == current_stage_id ) :
				valid_stage = True
				break;
			current_stage_index = current_stage_index + 1
	except NameError:
		print( 'stage_ids undefined\n' )
		exit( 1 )
	
	if ( not( valid_stage ) ) :
		print( 'Invalid stage specified %s' % current_stage_id )
		exit( 1 )
	
	if ( current_stage_index < 1 or current_stage_index > 4 ) :
		print( 'Nothing to run for this stage.  Exiting.\n')
		exit( 0 )

	# read in the stage-specific code
	
	
	stage_specific_fname = os.path.join( scripts_path, current_stage_id+'.py' )
	
	try:
		stage_specific = open( stage_specific_fname ).read()
	except:
		print( 'Could not open %s for reading.' % stage_specific_fname )  
		exit( 1 )
	
	exec( stage_specific )
	
	# copy the input files to the temporary directory for processing.
	copy_stage_inputs( working_dir_path )
	
	build_params_list()
	if ( use_xml ) :
		write_shapeworks_xml( os.path.join( working_dir_path, 'stage.xml' ) )
	else :
		write_shapeworks_params( os.path.join( working_dir_path, 'stage.params' ) )

	state_path = os.path.join( os.path.join( project_path, stage_id ), "state.txt" )

	# write the updated state
	try:
		if ( not reuse_previous_results ) :
			update_state( state_path )
	except: # reuse_previous_results may be undefined....
		pass

	#sys.exit(0)
